@@ -1,5 +1,5 @@
 /************************************************************************************
-   Copyright (C) 2013,2015 MariaDB Corporation AB
+   Copyright (C) 2013,2017 MariaDB Corporation AB
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -167,57 +167,50 @@ SQLRETURN MADB_DbcSetAttr(MADB_Dbc *Dbc, SQLINTEGER Attribute, SQLPOINTER ValueP
     break;
   case SQL_ATTR_AUTOCOMMIT:
     {
-      SQLINTEGER ValidAttrs[]= {2, SQL_AUTOCOMMIT_ON, SQL_AUTOCOMMIT_OFF};
-      /* (Here and in other places)We need to double-cast to avoid compiler warnings.
-         size_t is chosen as integer type that has the same size with pointer */
-      MADB_CHECK_ATTRIBUTE(Dbc, (SQLINTEGER)(size_t)ValuePtr, ValidAttrs);
+      SQLULEN ValidAttrs[]= {2, SQL_AUTOCOMMIT_ON, SQL_AUTOCOMMIT_OFF};
+      MADB_CHECK_ATTRIBUTE(Dbc, ValuePtr, ValidAttrs);
       /* if a connection is open, try to apply setting to the connection */
       if (Dbc->mariadb)
       {
         if (Dbc->EnlistInDtc) {
-          MADB_SetError(&Dbc->Error, MADB_ERR_25000, NULL, 0);
-          return SQL_ERROR;
+          return MADB_SetError(&Dbc->Error, MADB_ERR_25000, NULL, 0);
         }
         if (mysql_autocommit(Dbc->mariadb, (my_bool)(size_t)ValuePtr))
         {
-          MADB_SetError(&Dbc->Error, MADB_ERR_HY001, mysql_error(Dbc->mariadb), mysql_errno(Dbc->mariadb));
-          return SQL_ERROR;
+          return MADB_SetError(&Dbc->Error, MADB_ERR_HY001, mysql_error(Dbc->mariadb), mysql_errno(Dbc->mariadb));
         }
       }
-      Dbc->AutoCommit= (SQLUINTEGER)(size_t)ValuePtr;
+      Dbc->AutoCommit= (SQLUINTEGER)(SQLULEN)ValuePtr;
     }
     break;
   case SQL_ATTR_CONNECTION_DEAD:
     /* read only! */
-    MADB_SetError(&Dbc->Error, MADB_ERR_HY092, NULL, 0);
-    return Dbc->Error.ReturnValue;
-    break;
+    return MADB_SetError(&Dbc->Error, MADB_ERR_HY092, NULL, 0);
   case SQL_ATTR_CURRENT_CATALOG:
     {
       MADB_FREE(Dbc->CatalogName);
       if (isWChar)
         Dbc->CatalogName= MADB_ConvertFromWChar((SQLWCHAR *)ValuePtr, StringLength, NULL, &Dbc->charset, NULL);
       else
-        Dbc->CatalogName= my_strdup((char *)ValuePtr, MYF(0));
+        Dbc->CatalogName= _strdup((char *)ValuePtr);
 
       if (Dbc->mariadb &&
           mysql_select_db(Dbc->mariadb, Dbc->CatalogName))
       {
-        MADB_SetError(&Dbc->Error, MADB_ERR_HY001, mysql_error(Dbc->mariadb), mysql_errno(Dbc->mariadb));
-        return Dbc->Error.ReturnValue;
+        return MADB_SetError(&Dbc->Error, MADB_ERR_HY001, mysql_error(Dbc->mariadb), mysql_errno(Dbc->mariadb));
       }
     }
     break;
   case SQL_ATTR_LOGIN_TIMEOUT:
-    Dbc->LoginTimeout= (SQLUINTEGER)(size_t)ValuePtr;
+    Dbc->LoginTimeout= (SQLUINTEGER)(SQLULEN)ValuePtr;
     break;
   case SQL_ATTR_METADATA_ID:
-    Dbc->MetadataId= (SQLUINTEGER)(size_t)ValuePtr;
+    Dbc->MetadataId= (SQLUINTEGER)(SQLULEN)ValuePtr;
     break;
   case SQL_ATTR_ODBC_CURSORS:
     {
-      SQLINTEGER ValidAttrs[]= {3, SQL_CUR_USE_IF_NEEDED, SQL_CUR_USE_ODBC, SQL_CUR_USE_DRIVER};
-      MADB_CHECK_ATTRIBUTE(Dbc, (SQLULEN)ValuePtr, ValidAttrs);
+      SQLULEN ValidAttrs[]= {3, SQL_CUR_USE_IF_NEEDED, SQL_CUR_USE_ODBC, SQL_CUR_USE_DRIVER};
+      MADB_CHECK_ATTRIBUTE(Dbc, ValuePtr, ValidAttrs);
       if ((SQLULEN)ValuePtr != SQL_CUR_USE_ODBC)
         MADB_SetError(&Dbc->Error, MADB_ERR_01S02, NULL, 0);
       Dbc->OdbcCursors= SQL_CUR_USE_ODBC;
@@ -225,17 +218,14 @@ SQLRETURN MADB_DbcSetAttr(MADB_Dbc *Dbc, SQLINTEGER Attribute, SQLPOINTER ValueP
     break;
   case SQL_ATTR_ENLIST_IN_DTC:
     /* MS Distributed Transaction Coordinator not supported */
-    MADB_SetError(&Dbc->Error, MADB_ERR_HYC00, NULL, 0);
-    return Dbc->Error.ReturnValue;
-    break;
+    return MADB_SetError(&Dbc->Error, MADB_ERR_HYC00, NULL, 0);
   case SQL_ATTR_PACKET_SIZE:
     /* if connection was made, return HY001 */
     if (Dbc->mariadb)
     {
-      MADB_SetError(&Dbc->Error, MADB_ERR_HY001, NULL, 0);
-      return Dbc->Error.ReturnValue;
+      return MADB_SetError(&Dbc->Error, MADB_ERR_HY001, NULL, 0);
     }
-    Dbc->PacketSize= (SQLUINTEGER)(size_t)ValuePtr;
+    Dbc->PacketSize= (SQLUINTEGER)(SQLULEN)ValuePtr;
     break;
   case SQL_ATTR_QUIET_MODE:
     Dbc->QuietMode= (HWND)ValuePtr;
@@ -258,14 +248,13 @@ SQLRETURN MADB_DbcSetAttr(MADB_Dbc *Dbc, SQLINTEGER Attribute, SQLPOINTER ValueP
         if (MADB_IsolationLevel[i].SqlIsolation == (SQLLEN)ValuePtr)
         {
           char StmtStr[128];
-          my_snprintf(StmtStr, 128, "SET SESSION TRANSACTION ISOLATION LEVEL %s",
+          _snprintf(StmtStr, sizeof(StmtStr), "SET SESSION TRANSACTION ISOLATION LEVEL %s",
                       MADB_IsolationLevel[i].StrIsolation);
           LOCK_MARIADB(Dbc);
           if (mysql_query(Dbc->mariadb, StmtStr))
           {
             UNLOCK_MARIADB(Dbc);
-            MADB_SetError(&Dbc->Error, MADB_ERR_HY001, mysql_error(Dbc->mariadb), mysql_errno(Dbc->mariadb));
-            return Dbc->Error.ReturnValue;
+            return MADB_SetError(&Dbc->Error, MADB_ERR_HY001, mysql_error(Dbc->mariadb), mysql_errno(Dbc->mariadb));
           }
           UNLOCK_MARIADB(Dbc);
           ValidTx= TRUE;
@@ -274,11 +263,10 @@ SQLRETURN MADB_DbcSetAttr(MADB_Dbc *Dbc, SQLINTEGER Attribute, SQLPOINTER ValueP
       }
       if (!ValidTx)
       {
-        MADB_SetError(&Dbc->Error, MADB_ERR_HY024, NULL, 0);
-        return Dbc->Error.ReturnValue;
+        return MADB_SetError(&Dbc->Error, MADB_ERR_HY024, NULL, 0);
       }
     }
-    Dbc->TxnIsolation= (SQLLEN)ValuePtr;
+    Dbc->TxnIsolation= (SQLINTEGER)(SQLLEN)ValuePtr;
     break;
   default:
     break;
@@ -300,8 +288,7 @@ SQLRETURN MADB_DbcGetAttr(MADB_Dbc *Dbc, SQLINTEGER Attribute, SQLPOINTER ValueP
   if (Attribute == SQL_ATTR_CURRENT_CATALOG && !StringLengthPtr && 
       (!ValuePtr || !BufferLength))
   {
-    MADB_SetError(&Dbc->Error, MADB_ERR_01004, NULL, 0);
-    return Dbc->Error.ReturnValue;
+    return MADB_SetError(&Dbc->Error, MADB_ERR_01004, NULL, 0);
   }
 
   switch(Attribute) {
@@ -358,7 +345,11 @@ SQLRETURN MADB_DbcGetAttr(MADB_Dbc *Dbc, SQLINTEGER Attribute, SQLPOINTER ValueP
     MADB_SetError(&Dbc->Error, MADB_ERR_HYC00, NULL, 0);
     break;
   case SQL_ATTR_PACKET_SIZE:
-    *(SQLINTEGER *)ValuePtr= Dbc->PacketSize;
+    {
+      unsigned long packet_size= 0;
+      mysql_get_option(Dbc->mariadb, MYSQL_OPT_NET_BUFFER_LENGTH, &packet_size);
+      *(SQLINTEGER *)ValuePtr= (SQLINTEGER)packet_size;
+    }
     break;
   case SQL_ATTR_QUIET_MODE:
     Dbc->QuietMode= (HWND)ValuePtr;
@@ -434,22 +425,16 @@ MADB_Dbc *MADB_DbcInit(MADB_Env *Env)
   InitializeCriticalSection(&Connection->cs);
   /* Not sure that critical section is really needed here - this init routine is called when
      no one has the handle yet */
-  EnterCriticalSection(&Connection->cs);
+  EnterCriticalSection(&Connection->Environment->cs);
 
   /* Save connection in Environment list */
   Connection->ListItem.data= (void *)Connection;
-  Connection->Environment->Dbcs= list_add(Connection->Environment->Dbcs, &Connection->ListItem);
+  Connection->Environment->Dbcs= MADB_ListAdd(Connection->Environment->Dbcs, &Connection->ListItem);
 
-  LeaveCriticalSection(&Connection->cs);
+  LeaveCriticalSection(&Connection->Environment->cs);
 
   MADB_PutErrorPrefix(NULL, &Connection->Error);
 
-  if (!(Connection->mariadb= mysql_init(NULL)))
-  {
-    MADB_SetError(&Connection->Error, MADB_ERR_HY001, NULL, 0);
-    goto cleanup;
-  }
-
   return Connection;      
 cleanup:
   if (Connection)
@@ -477,13 +462,16 @@ SQLRETURN MADB_DbcFree(MADB_Dbc *Connection)
            more fingers movements
     LOCK_MARIADB(Dbc);*/
   if (Connection->mariadb)
+  {
     mysql_close(Connection->mariadb);
+    Connection->mariadb= NULL;
+  }
   /*UNLOCK_MARIADB(Dbc);*/
 
   /* todo: delete all descriptors */
 
   EnterCriticalSection(&Env->cs);
-  Connection->Environment->Dbcs= list_delete(Connection->Environment->Dbcs, &Connection->ListItem);
+  Connection->Environment->Dbcs= MADB_ListDelete(Connection->Environment->Dbcs, &Connection->ListItem);
   LeaveCriticalSection(&Env->cs);
 
   MADB_FREE(Connection->CatalogName);
@@ -501,23 +489,24 @@ SQLRETURN MADB_DbcFree(MADB_Dbc *Connection)
 SQLRETURN MADB_Dbc_GetCurrentDB(MADB_Dbc *Connection, SQLPOINTER CurrentDB, SQLINTEGER CurrentDBLength, 
                                 SQLSMALLINT *StringLengthPtr, my_bool isWChar) 
 {
-  SQLHANDLE Stmt;
+  MADB_Stmt *Stmt;
   SQLRETURN ret;
   SQLLEN Size;
   char Buffer[65 * sizeof(WCHAR)];
 
   MADB_CLEAR_ERROR(&Connection->Error);
-  ret= MA_SQLAllocHandle(SQL_HANDLE_STMT, (SQLHANDLE) Connection, &Stmt);
+  ret= MA_SQLAllocHandle(SQL_HANDLE_STMT, (SQLHANDLE) Connection, (SQLHANDLE*)&Stmt);
   if (!SQL_SUCCEEDED(ret))
     return ret;
-  if (!SQL_SUCCEEDED(MA_SQLExecDirect(Stmt, (SQLCHAR *)"SELECT IF(DATABASE() IS NOT NULL,DATABASE(),'null')", SQL_NTS)) ||
-      !SQL_SUCCEEDED(MA_SQLFetch(Stmt)))
+  if (!SQL_SUCCEEDED(Stmt->Methods->ExecDirect(Stmt, (SQLCHAR *)"SELECT IF(DATABASE() IS NOT NULL,DATABASE(),'null')", SQL_NTS)) ||
+      !SQL_SUCCEEDED(Stmt->Methods->Fetch(Stmt)))
   {
-    MADB_CopyError(&Connection->Error, &((MADB_Stmt *)Stmt)->Error);
+    MADB_CopyError(&Connection->Error, &Stmt->Error);
     goto end;
   }
   
-  ret= MA_SQLGetData(Stmt, 1, SQL_CHAR, Buffer, 65, &Size);
+  ret= Stmt->Methods->GetData(Stmt, 1, SQL_CHAR, Buffer, 65, &Size, TRUE);
+
  /* Size= (SQLINTEGER)MADB_SetString(isWChar ? Connection->CodePage : 0, CurrentDB, 
                       (isWChar) ? (int)(MIN(Size + sizeof(SQLWCHAR), CurrentDBLength) / sizeof(SQLWCHAR)) : 
                       (int)(MIN(Size + 1, CurrentDBLength)),
@@ -529,7 +518,7 @@ SQLRETURN MADB_Dbc_GetCurrentDB(MADB_Dbc *Connection, SQLPOINTER CurrentDB, SQLI
     *StringLengthPtr= isWChar ? (SQLSMALLINT)Size * sizeof(SQLWCHAR) : (SQLSMALLINT)Size;
   
 end:
-  MA_SQLFreeStmt(Stmt, SQL_CLOSE);
+  MA_SQLFreeStmt(Stmt, SQL_DROP);
   return Connection->Error.ReturnValue;
 }
 /* }}} */
@@ -632,12 +621,12 @@ SQLRETURN MADB_DbcConnectDB(MADB_Dbc *Connection,
   if (DSN_OPTION(Connection, MADB_OPT_FLAG_AUTO_RECONNECT))
     mysql_options(Connection->mariadb, MYSQL_OPT_RECONNECT, &my_reconnect);
 
-  if(DSN_OPTION(Connection, MADB_OPT_FLAG_NAMED_PIPE))
+  if (Dsn->IsNamedPipe) /* DSN_OPTION(Connection, MADB_OPT_FLAG_NAMED_PIPE) */
     mysql_options(Connection->mariadb, MYSQL_OPT_NAMED_PIPE, (void *)Dsn->ServerName);
 
   if (DSN_OPTION(Connection, MADB_OPT_FLAG_NO_SCHEMA))
     client_flags|= CLIENT_NO_SCHEMA;
-  if (DSN_OPTION(Connection, MADB_OPT_FLAG_PAD_SPACE))
+  if (DSN_OPTION(Connection, MADB_OPT_FLAG_IGNORE_SPACE))
     client_flags|= CLIENT_IGNORE_SPACE;
 
   if (DSN_OPTION(Connection, MADB_OPT_FLAG_FOUND_ROWS))
@@ -666,9 +655,15 @@ SQLRETURN MADB_DbcConnectDB(MADB_Dbc *Connection,
   }
   if (Dsn->SslVerify)
   {
-    const SQLPOINTER verify= (SQLPOINTER)1;
-    mysql_options(Connection->mariadb, MYSQL_OPT_SSL_VERIFY_SERVER_CERT, (const char*)verify);
+    const unsigned int verify= 0x01010101;
+    mysql_options(Connection->mariadb, MYSQL_OPT_SSL_VERIFY_SERVER_CERT, (const char*)&verify);
   }
+  else
+  {
+    const unsigned int verify= 0;
+    mysql_options(Connection->mariadb, MYSQL_OPT_SSL_VERIFY_SERVER_CERT, (const char*)&verify);
+  }
+
   if (!MADB_IS_EMPTY(Dsn->SslCrlPath))
   {
     mysql_options(Connection->mariadb, MYSQL_OPT_SSL_CRLPATH, Dsn->SslCrlPath);
@@ -681,16 +676,14 @@ SQLRETURN MADB_DbcConnectDB(MADB_Dbc *Connection,
     goto err;
   }
   
+  /* I guess it is better not to do that at all. Besides SQL_ATTR_PACKET_SIZE is actually not for max packet size */
   if (Connection->PacketSize)
   {
-    my_snprintf(StmtStr, 128, "SET max_allowed_packet_size=%ld", Connection->PacketSize);
+    /*_snprintf(StmtStr, 128, "SET GLOBAL max_allowed_packet=%ld", Connection-> PacketSize);
     if (mysql_query(Connection->mariadb, StmtStr))
-      goto err;
+      goto err;*/
   }
 
-  if (!Connection->CatalogName && Dsn->Catalog)
-    Connection->CatalogName= my_strdup(Dsn->Catalog, MYF(0));
-
   /* set default catalog */
   if (Connection->CatalogName && Connection->CatalogName[0])
   {
@@ -698,6 +691,9 @@ SQLRETURN MADB_DbcConnectDB(MADB_Dbc *Connection,
       goto err;
   }
 
+  if (!Connection->CatalogName && Dsn->Catalog)
+    Connection->CatalogName= _strdup(Dsn->Catalog);
+
   /* Turn sql_auto_is_null behavior off.
      For more details see: http://bugs.mysql.com/bug.php?id=47005 */
   if (mysql_query(Connection->mariadb, "SET SESSION SQL_AUTO_IS_NULL=0"))
@@ -713,7 +709,7 @@ SQLRETURN MADB_DbcConnectDB(MADB_Dbc *Connection,
     {
       if (MADB_IsolationLevel[i].SqlIsolation == Connection->IsolationLevel)
       {
-        my_snprintf(StmtStr, 128, "SET SESSION TRANSACTION ISOLATION LEVEL %s",
+        _snprintf(StmtStr, 128, "SET SESSION TRANSACTION ISOLATION LEVEL %s",
                     MADB_IsolationLevel[i].StrIsolation);
         if (mysql_query(Connection->mariadb, StmtStr))
           goto err;
@@ -721,6 +717,8 @@ SQLRETURN MADB_DbcConnectDB(MADB_Dbc *Connection,
       }
     }
 
+  MADB_SetCapabilities(Connection, mysql_get_server_version(Connection->mariadb));
+
   goto end;
 
 err:
@@ -937,6 +935,9 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
   case SQL_CONVERT_CHAR:
     MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, MADB_SUPPORTED_CONVERSIONS, StringLengthPtr);
     break;
+  case SQL_CONVERT_WCHAR:
+    MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, MADB_SUPPORTED_CONVERSIONS, StringLengthPtr);
+    break;
 #ifdef SQL_CONVERT_GUID
   case SQL_CONVERT_GUID:
     MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, 0, StringLengthPtr);
@@ -969,6 +970,9 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
   case SQL_CONVERT_LONGVARCHAR:
     MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, MADB_SUPPORTED_CONVERSIONS, StringLengthPtr);
     break;
+  case SQL_CONVERT_WLONGVARCHAR:
+    MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, MADB_SUPPORTED_CONVERSIONS, StringLengthPtr);
+    break;
   case SQL_CONVERT_NUMERIC:
     MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, MADB_SUPPORTED_CONVERSIONS, StringLengthPtr);
     break;
@@ -993,6 +997,9 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
   case SQL_CONVERT_VARCHAR:
     MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, MADB_SUPPORTED_CONVERSIONS, StringLengthPtr);
     break;
+  case SQL_CONVERT_WVARCHAR:
+    MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, MADB_SUPPORTED_CONVERSIONS, StringLengthPtr);
+    break;
   case SQL_CONVERT_FUNCTIONS:
     MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, 0, StringLengthPtr);
     break;
@@ -1054,14 +1061,13 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
     break;
   case SQL_DBMS_VER:
     {
-      char Version[11];
-      ulong ServerVersion= 0L;
-      unsigned int Major=0, Minor=0, Patch= 0;
+      char Version[13];
+      unsigned long ServerVersion= 0L;
       
       if (Dbc->mariadb)
       {
         ServerVersion= mysql_get_server_version(Dbc->mariadb);
-        my_snprintf(Version, 11, "%02u.%02u.%06u", ServerVersion / 10000,
+        _snprintf(Version, sizeof(Version), "%02u.%02u.%06u", ServerVersion / 10000,
                     (ServerVersion % 10000) / 100, ServerVersion % 100);
       }
       else
@@ -1103,7 +1109,7 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
   case SQL_DRIVER_NAME:
     SLen= (SQLSMALLINT)MADB_SetString(isWChar ? &Dbc->charset : NULL,
                                      (void *)InfoValuePtr, BUFFER_CHAR_LEN(BufferLength, isWChar), 
-                                     isWChar ? "maodbcu.dll" : "maodbca.dll", SQL_NTS, &Dbc->Error);
+                                     MADB_DRIVER_NAME, SQL_NTS, &Dbc->Error);
     break;
   case SQL_DRIVER_ODBC_VER:
     {
@@ -1176,7 +1182,7 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
                                      "Y", SQL_NTS, &Dbc->Error);
     break;
   case SQL_FILE_USAGE:
-    MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, SQL_FILE_NOT_SUPPORTED, StringLengthPtr);
+    MADB_SET_NUM_VAL(SQLUSMALLINT, InfoValuePtr, SQL_FILE_NOT_SUPPORTED, StringLengthPtr);
     break;
   case SQL_FORWARD_ONLY_CURSOR_ATTRIBUTES1:
     MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, SQL_CA1_ABSOLUTE |
@@ -1396,9 +1402,6 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
     SLen= (SQLSMALLINT)MADB_SetString(isWChar ? &Dbc->charset : NULL, (void *)InfoValuePtr, BUFFER_CHAR_LEN(BufferLength, isWChar), 
                                      "Y", SQL_NTS, &Dbc->Error);
     break;
-  case SQL_POS_OPERATIONS:
-    MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, 0, StringLengthPtr);
-    break;
   case SQL_QUOTED_IDENTIFIER_CASE:  
     MADB_SET_NUM_VAL(SQLUSMALLINT, InfoValuePtr, SQL_IC_SENSITIVE, StringLengthPtr);
     break;
@@ -1585,7 +1588,21 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
                                      BUFFER_CHAR_LEN(BufferLength, isWChar),
                                      "N", SQL_NTS, &Dbc->Error);
     break;
-  default:
+  /* 2.0 types */
+  case SQL_POS_OPERATIONS:
+    MADB_SET_NUM_VAL(SQLINTEGER, InfoValuePtr, SQL_POS_POSITION | SQL_POS_REFRESH | SQL_POS_UPDATE | SQL_POS_DELETE | SQL_POS_ADD,
+                     StringLengthPtr);
+    break;
+  case SQL_STATIC_SENSITIVITY:
+    MADB_SET_NUM_VAL(SQLINTEGER, InfoValuePtr, SQL_SS_DELETIONS | SQL_SS_UPDATES, StringLengthPtr);
+    break;
+  case SQL_LOCK_TYPES:
+    MADB_SET_NUM_VAL(SQLINTEGER, InfoValuePtr, SQL_LCK_NO_CHANGE, StringLengthPtr);
+    break;
+  case SQL_SCROLL_CONCURRENCY:
+    MADB_SET_NUM_VAL(SQLINTEGER, InfoValuePtr, SQL_SCCO_READ_ONLY | SQL_SCCO_OPT_VALUES, StringLengthPtr);
+    break;
+default:
     MADB_SetError(&Dbc->Error, MADB_ERR_HY096, NULL, 0);
     return Dbc->Error.ReturnValue;
   }
@@ -1634,34 +1651,37 @@ SQLRETURN MADB_DriverConnect(MADB_Dbc *Dbc, SQLHWND WindowHandle, SQLCHAR *InCon
 
   Dsn= MADB_DSN_Init();
 
-  if (!MADB_ParseDSNString(Dsn, (char *)InConnectionString, StringLength1, ';'))
+  if (!MADB_ReadConnString(Dsn, (char *)InConnectionString, StringLength1, ';'))
   {
     MADB_SetError(&Dbc->Error, MADB_ERR_HY000, "Error while parsing DSN", 0);
     goto error;
   }
 
-  /* Read DSN, but don't overwrite values! */
-  if (Dsn->DSNName)
-    MADB_ReadDSN(Dsn, NULL, FALSE);
-    
-  /* if DSN prompt is off, adjust DriverCompletion */
+  /* if DSN prompt is off, adjusting DriverCompletion */
   if (Dsn->ConnectPrompt)
     DriverCompletion= SQL_DRIVER_NOPROMPT;
 
   switch (DriverCompletion) {
-  case SQL_DRIVER_COMPLETE:
   case SQL_DRIVER_COMPLETE_REQUIRED:
+  case SQL_DRIVER_COMPLETE:
   case SQL_DRIVER_NOPROMPT:
     {
       SQLRETURN ret= MADB_DbcConnectDB(Dbc, Dsn);
       if (SQL_SUCCEEDED(ret))
+      {
         goto end;
-      else
+      }
+      else if (DriverCompletion == SQL_DRIVER_NOPROMPT)
+      {
+        /* For SQL_DRIVER_COMPLETE(_REQUIRED) this is not the end - will show prompt for user */
         goto error;
+      }
     }
+    /* If we got here, it means that we had unsuccessful connect attempt with SQL_DRIVER_COMPLETE(_REQUIRED) completion
+       Have to clean that error */
+    MADB_CLEAR_ERROR(&Dbc->Error);
     break;
   case SQL_DRIVER_PROMPT:
-    Dsn->isPrompt= MAODBC_PROMPT/*_REQUIRED*/;
     break;
   default:
     MADB_SetError(&Dbc->Error, MADB_ERR_HY110, NULL, 0);
@@ -1670,12 +1690,17 @@ SQLRETURN MADB_DriverConnect(MADB_Dbc *Dbc, SQLHWND WindowHandle, SQLCHAR *InCon
   }
 
   /* Without window handle we can't show a dialog */
-  if (DriverCompletion == SQL_DRIVER_NOPROMPT && !WindowHandle)
+  if (DriverCompletion != SQL_DRIVER_NOPROMPT && !WindowHandle)
   {
     MADB_SetError(&Dbc->Error, MADB_ERR_IM008, NULL, 0);
     goto error;
   }
   
+  if (DriverCompletion == SQL_DRIVER_COMPLETE_REQUIRED)
+    Dsn->isPrompt= MAODBC_PROMPT_REQUIRED;
+  else
+    Dsn->isPrompt= MAODBC_PROMPT;
+
   /* We need to obtain the driver name to load maodbcs.dll, if it's not stored inside DSN, 
      error IM007 (dialog prohibited) will be returned */
   if (!Dsn->Driver)
@@ -1717,16 +1742,34 @@ SQLRETURN MADB_DriverConnect(MADB_Dbc *Dbc, SQLHWND WindowHandle, SQLCHAR *InCon
     goto error;
 end:
   Dbc->Dsn= Dsn;
-  Length= MADB_DsnToString(Dsn, (char *)OutConnectionString, BufferLength);
+  /* Dialog returns bitmap - syncing corresponding properties */
+  MADB_DsnUpdateOptionsFields(Dsn);
+  if (Dsn->isPrompt)
+  {
+    /* If prompt/complete(_required), and dialog was succusefully showed - we generate string from the result DSN */
+    Length= MADB_DsnToString(Dsn, (char *)OutConnectionString, BufferLength);
+  }
+  else
+  {
+    if (StringLength1 == SQL_NTS)
+    {
+      StringLength1= (SQLSMALLINT)strlen(InConnectionString);
+    }
+    if (OutConnectionString && BufferLength)
+    {
+      /* Otherwise we are supposed to simply copy incoming connection string */
+      strncpy_s((char *)OutConnectionString, BufferLength, InConnectionString, StringLength1);
+    }
+    Length= StringLength1;
+  }
   if (StringLength2Ptr)
     *StringLength2Ptr= (SQLSMALLINT)Length;
+
   if (OutConnectionString && BufferLength && Length > BufferLength)
   {
     MADB_SetError(&Dbc->Error, MADB_ERR_01004, NULL, 0);
     return Dbc->Error.ReturnValue;
   }
-  if (StringLength2Ptr)
-    *StringLength2Ptr= Length;
   return ret;
 error:
   DSNPrompt_Free(&DSNPrompt);
