@@ -1,5 +1,5 @@
 /************************************************************************************
-   Copyright (C) 2013 SkySQL AB
+   Copyright (C) 2013, 2017 MariaDB Corporation AB
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -19,6 +19,8 @@
 #include <ma_odbc.h>
 #include <stdint.h>
 
+#define MADB_FIELD_IS_BINARY(_field) ((_field).charsetnr == BINARY_CHARSETNR)
+
 void CloseMultiStatements(MADB_Stmt *Stmt)
 {
   unsigned int i;
@@ -30,18 +32,23 @@ void CloseMultiStatements(MADB_Stmt *Stmt)
   }
   MADB_FREE(Stmt->MultiStmts);
   Stmt->MultiStmtCount= 0;
+  Stmt->stmt= NULL;
 }
 
 
 /* Required, but not sufficient condition */
 BOOL QueryIsPossiblyMultistmt(char *queryStr)
 {
-  if (strchr(queryStr, ';'))
+  char *semicolon_pos= strchr(queryStr, ';');
+  /* String supposed to come here trimmed. Checking that semicolon is not last char in the string */
+  if (semicolon_pos != NULL && semicolon_pos < queryStr + strlen(queryStr) - 1)
   {
     /* CREATE PROCEDURE uses semicolons but is not supported in prepared statement
         protocol */
     if (!MADB_IsStatementSupported(queryStr, "CREATE", "PROCEDURE"))
       return FALSE;
+    if (!MADB_IsStatementSupported(queryStr, "CREATE", "FUNCTION"))
+      return FALSE;
     if (!MADB_IsStatementSupported(queryStr, "CREATE", "DEFINER"))
       return FALSE;
     return TRUE;
@@ -50,64 +57,106 @@ BOOL QueryIsPossiblyMultistmt(char *queryStr)
   return FALSE;
 }
 
-unsigned int GetMultiStatements(MADB_Stmt *Stmt, char *StmtStr, size_t Length)
+#define STRING_OR_COMMENT() (quote[0] || quote[1] || comment)
+unsigned int GetMultiStatements(MADB_Stmt *Stmt, char *StmtStr, SQLINTEGER Length)
 {
-  char *p, *last, *prev= NULL;
+  char *p, *prev= NULL, wait_char= 0;
   unsigned int statements= 1;
   int quote[2]= {0,0}, comment= 0;
   char *end;
   MYSQL_STMT *stmt;
   char *StmtCopy= NULL;
 
+  /* mysql_stmt_init requires locking as well */
+  LOCK_MARIADB(Stmt->Connection);
   stmt= mysql_stmt_init(Stmt->Connection->mariadb);
 
   /* if the entire stmt string passes, we don't have multistatement */
   if (stmt && !mysql_stmt_prepare(stmt, StmtStr, Length))
   {
     mysql_stmt_close(stmt);
+    UNLOCK_MARIADB(Stmt->Connection);
     return 1;
   }
   mysql_stmt_close(stmt);
+
   /* make sure we don't have trailing whitespace or semicolon */
   if (Length)
   {
     end= StmtStr + Length - 1;
     while (end > StmtStr && (isspace(*end) || *end == ';'))
-      end--;
-    Length= end - StmtStr;
-  }
-  p= last= StmtCopy= my_strdup(StmtStr, MYF(0));
-  
-  while (p < StmtCopy + Length)
-  {
-    switch (*p) {
-    case ';':
-      if (!quote[0] && !quote[1] && !comment)
-      {
-        statements++;
-        last= p + 1;
-        *p= '\0';
-      }
-      break;
-    case '/':
-      if (!comment && (p < StmtCopy + Length + 1) && (char)*(p+1) ==  '*')
-        comment= 1;
-      else if (comment && (p > StmtCopy) && (char)*(p-1) == '*')
-        comment= 0;
-      break;
-    case '\"':
-      if (prev && *prev != '\\')
-        quote[0] = !quote[0];
-      break;
-    case 39:
-      if (prev && *prev != '\\')
-        quote[1] = !quote[1];
-      break;
-    default:
-      break;
+    {
+      --end;
+      --Length;
     }
+  }
+  /* We can have here not NULL-terminated string as a source, thus we need to allocate, copy meaningful characters and
+     add NULL */
+  p= StmtCopy= MADB_ALLOC(Length + 1);
+  strncpy(StmtCopy, StmtStr, Length);
+  StmtCopy[Length]= '\0';
+
+  end= StmtCopy + Length;
+  
+  while (p < end)
+  {
+    if (wait_char)
+    {
+      if (*p == wait_char && *prev != '\\')
+      {
+        wait_char= 0;
+      }
+    }
+    else
+    {
+      switch (*p) {
+      case '-':
+        if (!STRING_OR_COMMENT() && (p < end - 1) && (char)*(p+1) ==  '-')
+        {
+          wait_char= '\n';
+        }
+        break;
+      case '#':
+        if (!STRING_OR_COMMENT())
+        {
+          wait_char= '\n';
+        }
+        break;
+      case ';':
+        if (!STRING_OR_COMMENT())
+        {
+          statements++;
+          *p= '\0';
+        }
+        break;
+      case '/':
+        if (!STRING_OR_COMMENT() && (p < end - 1) && (char)*(p+1) ==  '*')
+          comment= 1;
+        else if (comment && (p > StmtCopy) && (char)*(prev) == '*')
+          comment= 0;
+        break;
+      case '"':
+        quote[0] = !STRING_OR_COMMENT();
+        break;
+      case '\'':
+        quote[1] = !STRING_OR_COMMENT();
+        break;
+      case '\\':
+        /* *end is \0, so we are safe here to increment by 2 bytes */
+        if ((Stmt->Connection->mariadb->server_status & SERVER_STATUS_NO_BACKSLASH_ESCAPES) == 0  && p < end - 1)
+        {
+          /* Skipping escaped character and resetting prev */
+          p+= 2;
+          prev= 0;
+          continue;
+        }
+      default:
+        break;
+      }
+    }
+
     prev= p;
-    p++;
+    ++p;
   }
 
   if (statements > 1)
@@ -124,14 +173,19 @@ unsigned int GetMultiStatements(MADB_Stmt *Stmt, char *StmtStr, size_t Length)
     {
       /* Need to be incremented before CloseMultiStatements() */
       ++Stmt->MultiStmtCount;
-      Stmt->MultiStmts[i]= mysql_stmt_init(Stmt->Connection->mariadb);
-      MDBUG_C_PRINT(Stmt->Connection, "-->inited&preparing %0x(%d)", Stmt->MultiStmts[i], i);
-      if (mysql_stmt_prepare(Stmt->MultiStmts[i], p, strlen(p)))
+      Stmt->MultiStmts[i]= i == 0 ? Stmt->stmt : mysql_stmt_init(Stmt->Connection->mariadb);
+      MDBUG_C_PRINT(Stmt->Connection, "-->inited&preparing %0x(%d,%s)", Stmt->MultiStmts[i], i, p);
+
+      if (mysql_stmt_prepare(Stmt->MultiStmts[i], p, (unsigned long)strlen(p)))
       {
         MADB_SetNativeError(&Stmt->Error, SQL_HANDLE_STMT, Stmt->MultiStmts[i]);
         CloseMultiStatements(Stmt);
         if (StmtCopy)
-          my_free(StmtCopy);
+        {
+          free(StmtCopy);
+        }
+        UNLOCK_MARIADB(Stmt->Connection);
+
         return 0;
       }
       if (mysql_stmt_param_count(Stmt->MultiStmts[i]) > MaxParams)
@@ -139,14 +193,17 @@ unsigned int GetMultiStatements(MADB_Stmt *Stmt, char *StmtStr, size_t Length)
       p+= strlen(p) + 1;
       ++i;
     }
+    UNLOCK_MARIADB(Stmt->Connection);
+
     if (MaxParams)
       Stmt->params= (MYSQL_BIND *)MADB_CALLOC(sizeof(MYSQL_BIND) * MaxParams);
   }
   if (StmtCopy)
-    my_free(StmtCopy);
+    free(StmtCopy);
 
   return statements;
 }
+#undef STRING_OR_COMMENT
 
 my_bool MADB_CheckPtrLength(SQLINTEGER MaxLength, char *Ptr, SQLINTEGER NameLen)
 {
@@ -173,38 +230,41 @@ int  MADB_GetWCharType(int Type)
 
 int MADB_KeyTypeCount(MADB_Dbc *Connection, char *TableName, int KeyFlag)
 {
-  int Count= 0;
+  int          Count= 0;
   unsigned int i;
-  char StmtStr[1024];
-  char *p= StmtStr;
-  char Database[65];
-  SQLHSTMT Stmt= NULL;
-  MADB_Stmt *KeyStmt;
+  char         StmtStr[1024];
+  char         *p= StmtStr;
+  char         Database[65];
+  MADB_Stmt    *Stmt= NULL;
+  MADB_Stmt    *KeyStmt;
   
-  MA_SQLGetConnectAttr((SQLHDBC)Connection, SQL_ATTR_CURRENT_CATALOG, Database, 65, NULL);
-  p+= my_snprintf(p, 1024, "SELECT * FROM ");
+  Connection->Methods->GetAttr(Connection, SQL_ATTR_CURRENT_CATALOG, Database, 65, NULL, FALSE);
+  p+= _snprintf(p, 1024, "SELECT * FROM ");
   if (Database)
-    p+= my_snprintf(p, 1024 - strlen(p), "`%s`.", Database);
-  p+= my_snprintf(p, 1024 - strlen(p), "%s LIMIT 0", TableName);
+  {
+    p+= _snprintf(p, 1024 - strlen(p), "`%s`.", Database);
+  }
+  p+= _snprintf(p, 1024 - strlen(p), "%s LIMIT 0", TableName);
   if (MA_SQLAllocHandle(SQL_HANDLE_STMT, (SQLHANDLE)Connection, (SQLHANDLE*)&Stmt) == SQL_ERROR ||
-      MA_SQLPrepare(Stmt, (SQLCHAR *)StmtStr, SQL_NTS) == SQL_ERROR ||
-      MA_SQLExecute(Stmt) == SQL_ERROR ||
-      MA_SQLFetch(Stmt) == SQL_ERROR)
-      goto end;
+    Stmt->Methods->ExecDirect(Stmt, (SQLCHAR *)StmtStr, SQL_NTS) == SQL_ERROR ||
+    Stmt->Methods->Fetch(Stmt) == SQL_ERROR)
+  {
+    goto end;
+  }
   KeyStmt= (MADB_Stmt *)Stmt;
   for (i=0; i < mysql_stmt_field_count(KeyStmt->stmt); i++)
     if (KeyStmt->stmt->fields[i].flags & KeyFlag)
       Count++;
 end:
   if (Stmt)
-    MA_SQLFreeStmt(Stmt, SQL_DROP);
+    Stmt->Methods->StmtFree(Stmt, SQL_DROP);
   return Count;
 }
 
 /* {{{ MADB_get_single_row */
 my_bool MADB_get_single_row(MADB_Dbc *Connection,
     const char *StmtString,
-    size_t Length,
+    SQLINTEGER Length,
     unsigned int NumCols,
     char **Buffers,
     size_t *Buffer_Lengths)
@@ -236,7 +296,7 @@ my_bool MADB_get_single_row(MADB_Dbc *Connection,
 /* }}} */
 
 /* {{{ MADB_CheckODBCType */
-bool MADB_CheckODBCType(SQLSMALLINT Type)
+BOOL MADB_CheckODBCType(SQLSMALLINT Type)
 {
   switch(Type)
   {
@@ -281,6 +341,20 @@ SQLSMALLINT MADB_GetTypeFromConciseType(SQLSMALLINT ConciseType)
   case SQL_C_TIME:
   case SQL_C_TIMESTAMP:
     return SQL_DATETIME;
+  case SQL_C_INTERVAL_YEAR:
+  case SQL_C_INTERVAL_YEAR_TO_MONTH:
+  case SQL_C_INTERVAL_MONTH:
+  case SQL_C_INTERVAL_DAY:
+  case SQL_C_INTERVAL_DAY_TO_HOUR:
+  case SQL_C_INTERVAL_DAY_TO_MINUTE:
+  case SQL_C_INTERVAL_DAY_TO_SECOND:
+  case SQL_C_INTERVAL_HOUR:
+  case SQL_C_INTERVAL_HOUR_TO_MINUTE:
+  case SQL_C_INTERVAL_HOUR_TO_SECOND:
+  case SQL_C_INTERVAL_MINUTE:
+  case SQL_C_INTERVAL_MINUTE_TO_SECOND:
+  case SQL_C_INTERVAL_SECOND:
+      return SQL_INTERVAL;
   default:
     return ConciseType;
   }
@@ -324,7 +398,7 @@ char *MADB_GetTypeName(MYSQL_FIELD Field)
     return "date";
   case MYSQL_TYPE_VARCHAR:
   case MYSQL_TYPE_VAR_STRING:
-    return (Field.flags & BINARY_FLAG) ? "varbinary" : "varchar";
+    return MADB_FIELD_IS_BINARY(Field) ? "varbinary" : "varchar";
   case MYSQL_TYPE_BIT:
     return "bit";
   case MYSQL_TYPE_ENUM:
@@ -332,15 +406,15 @@ char *MADB_GetTypeName(MYSQL_FIELD Field)
   case MYSQL_TYPE_SET:
     return "set";
   case MYSQL_TYPE_TINY_BLOB:
-    return (Field.flags & BINARY_FLAG) ? "tinyblob" : "tinytext";
+    return MADB_FIELD_IS_BINARY(Field) ? "tinyblob" : "tinytext";
   case MYSQL_TYPE_MEDIUM_BLOB:
-    return (Field.flags & BINARY_FLAG) ? "mediumblob" : "mediumtext";
+    return MADB_FIELD_IS_BINARY(Field) ? "mediumblob" : "mediumtext";
   case MYSQL_TYPE_LONG_BLOB:
-    return (Field.flags & BINARY_FLAG) ? "longblob" : "longtext";
+    return MADB_FIELD_IS_BINARY(Field) ? "longblob" : "longtext";
   case MYSQL_TYPE_BLOB:
-    return (Field.flags & BINARY_FLAG) ? "blob" : "text";
+    return MADB_FIELD_IS_BINARY(Field) ? "blob" : "text";
   case MYSQL_TYPE_STRING:
-    return (Field.flags & BINARY_FLAG) ? "binary" : "char";
+    return MADB_FIELD_IS_BINARY(Field) ? "binary" : "char";
   case MYSQL_TYPE_GEOMETRY:
     return "geometry";
   default:
@@ -351,25 +425,33 @@ char *MADB_GetTypeName(MYSQL_FIELD Field)
 
 MYSQL_RES *MADB_GetDefaultColumnValues(MADB_Stmt *Stmt, MYSQL_FIELD *fields)
 {
-  DYNAMIC_STRING DynStr;
+  MADB_DynString DynStr;
   unsigned int i;
   MYSQL_RES *result= NULL;
   
-  init_dynamic_string(&DynStr, "SELECT COLUMN_NAME, COLUMN_DEFAULT FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA='", 512, 512);
-  if (dynstr_append(&DynStr, fields[0].db) ||
-      dynstr_append(&DynStr, "' AND TABLE_NAME='") ||
-      dynstr_append(&DynStr, fields[0].org_table) ||
-      dynstr_append(&DynStr, "' AND COLUMN_NAME IN ("))
+  MADB_InitDynamicString(&DynStr, "SELECT COLUMN_NAME, COLUMN_DEFAULT FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA='", 512, 512);
+  if (MADB_DynstrAppend(&DynStr, fields[0].db) ||
+      MADB_DynstrAppend(&DynStr, "' AND TABLE_NAME='") ||
+      MADB_DynstrAppend(&DynStr, fields[0].org_table) ||
+      MADB_DynstrAppend(&DynStr, "' AND COLUMN_NAME IN ("))
     goto error;
 
   for (i=0; i < mysql_stmt_field_count(Stmt->stmt); i++)
   {
-    if (dynstr_append(&DynStr, i > 0 ? ",'" : "'") ||
-        dynstr_append(&DynStr, fields[i].org_name) ||
-        dynstr_append(&DynStr, "'"))
+    MADB_DescRecord *Rec= MADB_DescGetInternalRecord(Stmt->Ard, i, MADB_DESC_READ);
+
+    if (!Rec->inUse || MADB_ColumnIgnoredInAllRows(Stmt->Ard, Rec) == TRUE)
+    {
+      continue;
+    }
+    if (MADB_DynstrAppend(&DynStr, i > 0 ? ",'" : "'") ||
+      MADB_DynstrAppend(&DynStr, fields[i].org_name) ||
+      MADB_DynstrAppend(&DynStr, "'"))
+    {
       goto error;
+    }
   }
-  if (dynstr_append(&DynStr, ") AND COLUMN_DEFAULT IS NOT NULL"))
+  if (MADB_DynstrAppend(&DynStr, ") AND COLUMN_DEFAULT IS NOT NULL"))
     goto error;
 
   LOCK_MARIADB(Stmt->Connection);
@@ -379,7 +461,7 @@ MYSQL_RES *MADB_GetDefaultColumnValues(MADB_Stmt *Stmt, MYSQL_FIELD *fields)
   
 error:
     UNLOCK_MARIADB(Stmt->Connection);
-    dynstr_free(&DynStr);
+    MADB_DynstrFree(&DynStr);
     return result;
 }
 
@@ -387,7 +469,7 @@ char *MADB_GetDefaultColumnValue(MYSQL_RES *res, const char *Column)
 {
   MYSQL_ROW row;
 
-  if (!res->row_count)
+  if (res == NULL || !res->row_count)
     return NULL;
   mysql_data_seek(res, 0);
   while ((row= mysql_fetch_row(res)))
@@ -398,7 +480,7 @@ char *MADB_GetDefaultColumnValue(MYSQL_RES *res, const char *Column)
   return NULL;
 }
 
-size_t MADB_GetDataSize(MADB_DescRecord *Record, MYSQL_FIELD Field, CHARSET_INFO *charset)
+size_t MADB_GetDataSize(MADB_DescRecord *Record, MYSQL_FIELD Field, MARIADB_CHARSET_INFO *charset)
 {
   switch(Record->ConciseType)
   {
@@ -428,8 +510,7 @@ size_t MADB_GetDataSize(MADB_DescRecord *Record, MYSQL_FIELD Field, CHARSET_INFO
     return Field.length;
   default:
     {
-      if (Field.flags & BINARY_FLAG || Field.charsetnr == BINARY_CHARSETNR
-        || charset == NULL || charset->char_maxlen < 2/*i.e.0||1*/)
+      if (MADB_FIELD_IS_BINARY(Field) || charset == NULL || charset->char_maxlen < 2/*i.e.0||1*/)
       {
         return Field.length;
       }
@@ -442,7 +523,7 @@ size_t MADB_GetDataSize(MADB_DescRecord *Record, MYSQL_FIELD Field, CHARSET_INFO
 }
 
 /* {{{ MADB_GetDisplaySize */
-size_t MADB_GetDisplaySize(MYSQL_FIELD Field, CHARSET_INFO *charset)
+size_t MADB_GetDisplaySize(MYSQL_FIELD Field, MARIADB_CHARSET_INFO *charset)
 {
   /* Todo: check these values with output from mysql --with-columntype-info */
   switch (Field.type) {
@@ -487,7 +568,7 @@ size_t MADB_GetDisplaySize(MYSQL_FIELD Field, CHARSET_INFO *charset)
   case MYSQL_TYPE_VARCHAR:
   case MYSQL_TYPE_VAR_STRING:
   {
-    if (Field.flags & BINARY_FLAG || Field.charsetnr == BINARY_CHARSETNR)
+    if (MADB_FIELD_IS_BINARY(Field))
     {
       return Field.length*2; /* ODBC specs says we should give 2 characters per byte to display binaray data in hex form */
     }
@@ -509,7 +590,7 @@ size_t MADB_GetDisplaySize(MYSQL_FIELD Field, CHARSET_INFO *charset)
 /* {{{ MADB_GetOctetLength */
 size_t MADB_GetOctetLength(MYSQL_FIELD Field, unsigned short MaxCharLen)
 {
-  size_t Length= MIN(INT_MAX32,Field.length);
+  size_t Length= MIN(MADB_INT_MAX32, Field.length);
 
   switch (Field.type) {
   case MYSQL_TYPE_NULL:
@@ -649,13 +730,13 @@ SQLSMALLINT MADB_GetODBCType(MYSQL_FIELD *field)
     case MYSQL_TYPE_BLOB:
     case MYSQL_TYPE_MEDIUM_BLOB:
     case MYSQL_TYPE_LONG_BLOB:
-      return field->flags & BINARY_FLAG ? SQL_LONGVARBINARY : SQL_LONGVARCHAR;
+      return MADB_FIELD_IS_BINARY(*field) ? SQL_LONGVARBINARY : SQL_LONGVARCHAR;
     case MYSQL_TYPE_LONGLONG:
       return SQL_BIGINT;
-     case MYSQL_TYPE_STRING:
-      return field->flags & BINARY_FLAG ? SQL_BINARY : SQL_CHAR;
+    case MYSQL_TYPE_STRING:
+      return MADB_FIELD_IS_BINARY(*field) ? SQL_BINARY : SQL_CHAR;
     case MYSQL_TYPE_VAR_STRING:
-      return field->flags & BINARY_FLAG ? SQL_VARBINARY : SQL_VARCHAR;
+      return MADB_FIELD_IS_BINARY(*field) ? SQL_VARBINARY : SQL_VARCHAR;
     case MYSQL_TYPE_SET:
     case MYSQL_TYPE_ENUM:
       return SQL_CHAR;
@@ -690,7 +771,7 @@ size_t MADB_GetTypeLength(SQLINTEGER SqlDataType, size_t Length)
     return sizeof(SQLINTEGER);
   case SQL_C_UBIGINT:
   case SQL_C_SBIGINT:
-    return sizeof(longlong);
+    return sizeof(long long);
   case SQL_C_DOUBLE:
     return sizeof(SQLDOUBLE);
   case SQL_C_FLOAT:
@@ -712,8 +793,8 @@ size_t MADB_GetTypeLength(SQLINTEGER SqlDataType, size_t Length)
 }
 /* }}} */
 
-/* {{{ MADB_GetTypeAndLength */
-int MADB_GetTypeAndLength(SQLINTEGER SqlDataType, my_bool *Unsigned, unsigned long *Length)
+/* {{{ MADB_GetMaDBTypeAndLength */
+int MADB_GetMaDBTypeAndLength(SQLINTEGER SqlDataType, my_bool *Unsigned, unsigned long *Length)
 {
   *Unsigned= 0;
   switch(SqlDataType)
@@ -724,13 +805,17 @@ int MADB_GetTypeAndLength(SQLINTEGER SqlDataType, my_bool *Unsigned, unsigned lo
   case SQL_C_UTINYINT:
     *Length= 1;
     *Unsigned= (SqlDataType == SQL_C_UTINYINT);
+
     return MYSQL_TYPE_TINY;
+
   case SQL_C_SHORT:
   case SQL_C_SSHORT:
   case SQL_C_USHORT:
     *Length= 2;
     *Unsigned= (SqlDataType == SQL_C_USHORT);
-     return MYSQL_TYPE_SHORT;
+
+    return MYSQL_TYPE_SHORT;
+ 
   case SQL_C_LONG:
   case SQL_C_SLONG:
   case SQL_C_ULONG:
@@ -739,7 +824,7 @@ int MADB_GetTypeAndLength(SQLINTEGER SqlDataType, my_bool *Unsigned, unsigned lo
     return MYSQL_TYPE_LONG;
   case SQL_C_UBIGINT:
   case SQL_C_SBIGINT:
-    *Length= sizeof(longlong);
+    *Length= sizeof(long long);
     *Unsigned= (SqlDataType == SQL_C_UBIGINT);
     return MYSQL_TYPE_LONGLONG;
   case SQL_C_DOUBLE:
@@ -763,6 +848,10 @@ int MADB_GetTypeAndLength(SQLINTEGER SqlDataType, my_bool *Unsigned, unsigned lo
   case SQL_C_TIMESTAMP:
     *Length= sizeof(SQL_TIMESTAMP_STRUCT);
     return MYSQL_TYPE_TIMESTAMP;
+  case SQL_C_INTERVAL_HOUR_TO_MINUTE:
+  case SQL_C_INTERVAL_HOUR_TO_SECOND:
+    *Length= sizeof(SQL_INTERVAL_STRUCT);
+    return MYSQL_TYPE_TIME;
   case SQL_C_CHAR:
     return MYSQL_TYPE_STRING;
   default:
@@ -771,7 +860,7 @@ int MADB_GetTypeAndLength(SQLINTEGER SqlDataType, my_bool *Unsigned, unsigned lo
 }
 /* }}} */
 
-void MADB_CopyMadbTimestamp(MYSQL_TIME *tm, MADB_Desc *Ard, MADB_DescRecord *ArdRecord, int Type, unsigned long RowNumber)
+SQLRETURN MADB_CopyMadbTimestamp(MADB_Stmt *Stmt, MYSQL_TIME *tm, MADB_Desc *Ard, MADB_DescRecord *ArdRecord, int Type, unsigned long RowNumber)
 {
   void *DataPtr= GetBindOffset(Ard, ArdRecord, ArdRecord->DataPtr, RowNumber, ArdRecord->OctetLength);
 
@@ -806,15 +895,17 @@ void MADB_CopyMadbTimestamp(MYSQL_TIME *tm, MADB_Desc *Ard, MADB_DescRecord *Ard
     }
     break;
     case SQL_C_TIME:
-    case SQL_TYPE_TIME:
+    case SQL_C_TYPE_TIME:
     {
       SQL_TIME_STRUCT *ts= (SQL_TIME_STRUCT *)DataPtr;
+
+      if (tm->hour > 23 || tm->minute > 59 || tm->second > 59)
+      {
+        return MADB_SetError(&Stmt->Error, MADB_ERR_22007, NULL, 0);
+      }
       ts->hour= tm->hour;
       ts->minute= tm->minute;
       ts->second= tm->second;
-       if (ts->hour + ts->minute + ts->second == 0)
-        if (ArdRecord->IndicatorPtr)
-          *ArdRecord->IndicatorPtr= SQL_NULL_DATA;
     }
     break;
     case SQL_C_DATE:
@@ -830,9 +921,11 @@ void MADB_CopyMadbTimestamp(MYSQL_TIME *tm, MADB_Desc *Ard, MADB_DescRecord *Ard
     }
     break;
   }
+
+  return SQL_SUCCESS;
 }
 
-void *GetBindOffset(MADB_Desc *Desc, MADB_DescRecord *Record, void *Ptr, unsigned long RowNumber, size_t PtrSize)
+void *GetBindOffset(MADB_Desc *Desc, MADB_DescRecord *Record, void *Ptr, SQLULEN RowNumber, size_t PtrSize)
 {
   size_t BindOffset= 0;
 
@@ -850,6 +943,25 @@ void *GetBindOffset(MADB_Desc *Desc, MADB_DescRecord *Record, void *Ptr, unsigne
   return (char *)Ptr + BindOffset;
 }
 
+/* Checking if column ignored in all bound rows. Should hel*/
+BOOL MADB_ColumnIgnoredInAllRows(MADB_Desc *Desc, MADB_DescRecord *Rec)
+{
+  SQLULEN row;
+  SQLLEN *IndicatorPtr;
+
+  for (row= 0; row < Desc->Header.ArraySize; ++row)
+  {
+    IndicatorPtr= (SQLLEN *)GetBindOffset(Desc, Rec, Rec->IndicatorPtr, row, sizeof(SQLLEN));
+
+    if (IndicatorPtr == NULL || *IndicatorPtr != SQL_COLUMN_IGNORE)
+    {
+      return FALSE;
+    }
+  }
+
+  return TRUE;
+}
+
 
 void MADB_NumericInit(SQL_NUMERIC_STRUCT *number, MADB_DescRecord *Ard)
 {
@@ -903,7 +1015,7 @@ int MADB_CharToSQLNumeric(char *buffer, MADB_Desc *Ard, MADB_DescRecord *ArdReco
     {
       short digits_total= 0, 
             digits_significant= 0;
-      digits_count= dot - p;
+      digits_count= (short)(dot - p);
       memcpy(digits, p, digits_count);
       p= dot + 1;
       while (*p)
@@ -944,7 +1056,7 @@ int MADB_CharToSQLNumeric(char *buffer, MADB_Desc *Ard, MADB_DescRecord *ArdReco
       {
         return MADB_ERR_22003;
       }
-      digits_count= p - start;
+      digits_count= (short)(p - start);
       memcpy(digits, start, digits_count);
       number->scale= ArdRecord->Scale ? ArdRecord->Scale : 0;
     }
@@ -963,7 +1075,7 @@ int MADB_CharToSQLNumeric(char *buffer, MADB_Desc *Ard, MADB_DescRecord *ArdReco
       if (OldVal != Val)
         return MADB_ERR_22003;
       _snprintf(digits, sizeof(digits), "%lld", Val);
-      digits_count= strlen(digits);
+      digits_count= (short)strlen(digits);
       if (digits_count > number->precision)
         return MADB_ERR_22003;
     }
@@ -1000,98 +1112,6 @@ int MADB_CharToSQLNumeric(char *buffer, MADB_Desc *Ard, MADB_DescRecord *ArdReco
   return ret;
 }
 
-/* {{{ MADB_SqlNumericToChar */
-size_t MADB_SqlNumericToChar(SQL_NUMERIC_STRUCT *Numeric, char *Buffer, int *ErrorCode)
-{
-  longlong Numerator= 0;
-  longlong Denominator= 1;
-  ulonglong Left= 0, Right= 0;
-  int Scale= 0;
-  int ppos= 0;
-  long ByteDenominator= 1;
-  int i;
-  char *p;
-  my_bool hasDot= FALSE;
-
-  Buffer[0]= 0;
-  *ErrorCode= 0;
-
-  Scale+= (Numeric->scale < 0) ? -Numeric->scale : Numeric->scale;
-      
-  for (i=0; i < SQL_MAX_NUMERIC_LEN; ++i)
-  {
-    Numerator+= Numeric->val[i] * ByteDenominator;
-    ByteDenominator<<= 8;
-  }
-  if (!Numeric->sign)
-    Numerator= -Numerator;
-  Denominator= (longlong)pow(10, Scale);
-  Left= Numerator / Denominator;
-  //_i64toa_s(Numerator, Buffer, 38, 10);
-  if (Numeric->scale > 0)
-  {
-    char tmp[38];
-    _snprintf(tmp,38, "%%%d.%df", Numeric->precision, Numeric->scale);
-    _snprintf(Buffer, 38, tmp, Numerator / pow(10, Scale));
-  }
-  else
-  {
-    _snprintf(Buffer, 38, "%lld", Numerator);
-    while (strlen(Buffer) < (size_t)(Numeric->precision - Numeric->scale))
-      strcat(Buffer, "0");
-  }
-  
-  
-  if (Buffer[0] == '-')
-    Buffer++;
- 
- /* Truncation checks:
-     1st ensure, that the digits before decimal point will fit */
-  if ((p= strchr(Buffer, '.')))
-  {
-    if (p - Buffer - 1 > Numeric->precision)
-    {
-      *ErrorCode= MADB_ERR_22003;
-      Buffer[Numeric->precision]= 0;
-      goto end;
-    }
-    if (Numeric->scale > 0 && Left > 0 && (p - Buffer) + strlen(p) > Numeric->precision)
-    {
-      *ErrorCode= MADB_ERR_01S07;
-      Buffer[Numeric->precision + 1]= 0;
-      goto end;
-    }
-  }
-  while (Numeric->scale < 0 && strlen(Buffer) < (size_t)(Numeric->precision - Numeric->scale))
-    strcat(Buffer, "0");
-    
-
-  if (strlen(Buffer) > (size_t)(Numeric->precision + Scale) && Numeric->scale > 0)
-    *ErrorCode= MADB_ERR_01S07;
-
-end:
-    /* check if last char is decimal point */
-  if (strlen(Buffer) && Buffer[strlen(Buffer)-1] == '.')
-    Buffer[strlen(Buffer)-1] = 0;
-  if (!Numeric->sign)
-    Buffer--;
-  return strlen(Buffer);
-}
-/* }}} */
-
-/* {{{ trim */
-char *trim(char *Str)
-{
-  char *end;
-  while (Str && iswspace(Str[0]))
-    Str++;
-  end= Str + strlen(Str) - 1;
-  while (iswspace(*end))
-    *end--= 0;
-  return Str;
-}
-/* }}} */
-
 /* {{{ MADB_GetHexString */
 size_t MADB_GetHexString(char *BinaryBuffer, size_t BinaryLength,
                           char *HexBuffer, size_t HexLength)
@@ -1130,16 +1150,17 @@ unsigned long MADB_StmtDataTell(MADB_Stmt *Stmt)
 
 SQLRETURN MADB_DaeStmt(MADB_Stmt *Stmt, SQLUSMALLINT Operation)
 {
-  char *TableName= MADB_GetTableName(Stmt);
-  char *CatalogName= MADB_GetCatalogName(Stmt);
-  DYNAMIC_STRING DynStmt;
+  char          *TableName=   MADB_GetTableName(Stmt);
+  char          *CatalogName= MADB_GetCatalogName(Stmt);
+  MADB_DynString DynStmt;
 
   MADB_CLEAR_ERROR(&Stmt->Error);
-  memset(&DynStmt, 0, sizeof(DYNAMIC_STRING));
+  memset(&DynStmt, 0, sizeof(MADB_DynString));
 
   if (Stmt->DaeStmt)
     Stmt->Methods->StmtFree(Stmt->DaeStmt, SQL_DROP);
   Stmt->DaeStmt= NULL;
+
   if (!SQL_SUCCEEDED(MA_SQLAllocHandle(SQL_HANDLE_STMT, (SQLHANDLE)Stmt->Connection, (SQLHANDLE *)&Stmt->DaeStmt)))
   {
     MADB_CopyError(&Stmt->Error, &Stmt->Connection->Error);
@@ -1149,53 +1170,76 @@ SQLRETURN MADB_DaeStmt(MADB_Stmt *Stmt, SQLUSMALLINT Operation)
   switch(Operation)
   {
   case SQL_ADD:
-    if (init_dynamic_string(&DynStmt, "INSERT INTO ", 1024, 1024) ||
+    if (MADB_InitDynamicString(&DynStmt, "INSERT INTO ", 1024, 1024) ||
         MADB_DynStrAppendQuoted(&DynStmt, CatalogName) ||
-        dynstr_append(&DynStmt, ".") ||
+        MADB_DynstrAppend(&DynStmt, ".") ||
         MADB_DynStrAppendQuoted(&DynStmt, TableName)||
         MADB_DynStrUpdateSet(Stmt, &DynStmt))
     {
-      dynstr_free(&DynStmt);
+      MADB_DynstrFree(&DynStmt);
       return Stmt->Error.ReturnValue;
     }
     Stmt->DataExecutionType= MADB_DAE_ADD;
     break;
   case SQL_DELETE:
-    if (init_dynamic_string(&DynStmt, "DELETE FROM ", 1024, 1024) ||
+    if (MADB_InitDynamicString(&DynStmt, "DELETE FROM ", 1024, 1024) ||
         MADB_DynStrAppendQuoted(&DynStmt, CatalogName) ||
-        dynstr_append(&DynStmt, ".") ||
+        MADB_DynstrAppend(&DynStmt, ".") ||
         MADB_DynStrAppendQuoted(&DynStmt, TableName) ||
         MADB_DynStrGetWhere(Stmt, &DynStmt, TableName, FALSE))
     {
-      dynstr_free(&DynStmt);
+      MADB_DynstrFree(&DynStmt);
       return Stmt->Error.ReturnValue;
     }
     Stmt->DataExecutionType= MADB_DAE_DELETE;
     break;
   case SQL_UPDATE:
-    Stmt->Methods->RefreshRowPtrs(Stmt);
-    if (init_dynamic_string(&DynStmt, "UPDATE ", 1024, 1024) ||
+    if (MADB_InitDynamicString(&DynStmt, "UPDATE ", 1024, 1024) ||
         MADB_DynStrAppendQuoted(&DynStmt, CatalogName) ||
-        dynstr_append(&DynStmt, ".") ||
+        MADB_DynstrAppend(&DynStmt, ".") ||
         MADB_DynStrAppendQuoted(&DynStmt, TableName)||
         MADB_DynStrUpdateSet(Stmt, &DynStmt)||
         MADB_DynStrGetWhere(Stmt, &DynStmt, TableName, FALSE))
     {
-      dynstr_free(&DynStmt);
+      MADB_DynstrFree(&DynStmt);
       return Stmt->Error.ReturnValue;
     }
     Stmt->DataExecutionType= MADB_DAE_UPDATE;
     break;
   }
   
-  if (!SQL_SUCCEEDED(MA_SQLPrepare(Stmt->DaeStmt, (SQLCHAR *)DynStmt.str, SQL_NTS)))
+  if (!SQL_SUCCEEDED(Stmt->DaeStmt->Methods->Prepare(Stmt->DaeStmt, (SQLCHAR *)DynStmt.str, SQL_NTS, FALSE)))
   {
     MADB_CopyError(&Stmt->Error, &Stmt->DaeStmt->Error);
     Stmt->Methods->StmtFree(Stmt->DaeStmt, SQL_DROP);
   }
    
 end:
-  dynstr_free(&DynStmt);
+  MADB_DynstrFree(&DynStmt);
   return Stmt->Error.ReturnValue;
 
 }
+
+int MADB_FindNextDaeParam(MADB_Desc *Desc, int InitialParam, SQLSMALLINT RowNumber)
+{
+  int             i;
+  MADB_DescRecord *Record;
+
+  for (i= InitialParam > -1 ? InitialParam + 1 : 0; i < Desc->Header.Count; i++)
+  {
+    if ((Record= MADB_DescGetInternalRecord(Desc, i, MADB_DESC_READ)))
+    {
+      if (Record->OctetLengthPtr)
+      {
+        /* Stmt->DaeRowNumber is 1 based */
+        SQLLEN *OctetLength = (SQLLEN *)GetBindOffset(Desc, Record, Record->OctetLengthPtr, RowNumber > 1 ? RowNumber - 1 : 0, sizeof(SQLLEN));
+        if (PARAM_IS_DAE(OctetLength))
+        {
+          return i;
+        }
+      }
+    }
+  }
+
+  return MADB_NOPARAM;
+}
