@@ -1,6 +1,6 @@
 /************************************************************************************
-   Copyright (C) 2013,2105 MariaDB Corporation AB
-
+   Copyright (C) 2013,2016 MariaDB Corporation AB
+   
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
@@ -110,12 +110,6 @@ SQLRETURN SQL_API SQLAllocConnect(SQLHANDLE InputHandle,
 /* }}} */
 
 /* {{{ SQLAllocStmt */
-SQLRETURN MA_SQLAllocStmt(SQLHANDLE InputHandle,
-                          SQLHANDLE *OutputHandlePtr)
-{
-  return MA_SQLAllocHandle(SQL_HANDLE_STMT, InputHandle, OutputHandlePtr);
-}
-
 SQLRETURN SQL_API SQLAllocStmt(SQLHANDLE InputHandle,
                                SQLHANDLE *OutputHandlePtr)
 {
@@ -123,7 +117,7 @@ SQLRETURN SQL_API SQLAllocStmt(SQLHANDLE InputHandle,
   MDBUG_C_DUMP(InputHandle, InputHandle, 0x);
   MDBUG_C_DUMP(InputHandle, OutputHandlePtr, 0x);
 
-  return MA_SQLAllocStmt(InputHandle, OutputHandlePtr);
+  return MA_SQLAllocHandle(SQL_HANDLE_STMT, InputHandle, OutputHandlePtr);
 }
 /* }}} */
 
@@ -263,7 +257,6 @@ SQLRETURN SQL_API SQLBulkOperations(SQLHSTMT StatementHandle,
   if (!Stmt)
     return SQL_INVALID_HANDLE;
   MADB_CLEAR_ERROR(&Stmt->Error);
-  Stmt->FetchType= MADB_FETCH_TYPE_BULK;
 
   MDBUG_C_ENTER(Stmt->Connection, "SQLBulkOperations");
   MDBUG_C_DUMP(Stmt->Connection, Stmt, 0x);
@@ -313,7 +306,7 @@ SQLRETURN MA_SQLCancel(SQLHSTMT StatementHandle)
       goto end;
     }
     
-    my_snprintf(StmtStr, 30, "KILL QUERY %ld", mysql_thread_id(Kill));
+    _snprintf(StmtStr, 30, "KILL QUERY %ld", mysql_thread_id(Kill));
     if (mysql_query(MariaDb, StmtStr))
     {
       mysql_close(MariaDb);
@@ -420,7 +413,7 @@ SQLRETURN SQL_API SQLColAttribute (SQLHSTMT StatementHandle,
 
   MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
 }
-
+/* }}} */
 
 /* {{{ SQLColAttributeW */
 SQLRETURN SQL_API SQLColAttributeW (SQLHSTMT StatementHandle,
@@ -459,28 +452,6 @@ SQLRETURN SQL_API SQLColAttributeW (SQLHSTMT StatementHandle,
 }
 /* }}} */
 
-/* {{{ MapColAttributesDescType */
-SQLUSMALLINT MapColAttributeDescType(SQLUSMALLINT FieldIdentifier)
-{
-  /* we need to map the old field identifiers, see bug ODBC-8 */
-  switch (FieldIdentifier)
-  {
-  case SQL_COLUMN_SCALE:
-    return SQL_DESC_SCALE;
-  case SQL_COLUMN_PRECISION:
-    return SQL_DESC_PRECISION;
-  case SQL_COLUMN_NULLABLE:
-    return SQL_DESC_NULLABLE;
-  case SQL_COLUMN_LENGTH:
-    return SQL_DESC_OCTET_LENGTH;
-  case SQL_COLUMN_NAME:
-    return SQL_DESC_NAME;
-  default:
-    return FieldIdentifier;
-  }
-}
-/* }}} */
-
 /* {{{ SQLColAttributes */
 SQLRETURN SQL_API SQLColAttributes(SQLHSTMT hstmt, 
 	SQLUSMALLINT icol,
@@ -639,9 +610,9 @@ SQLRETURN SQL_API SQLColumnsW(SQLHSTMT StatementHandle,
   MDBUG_C_ENTER(Stmt->Connection, "SQLColumns");
 
   CpCatalog= MADB_ConvertFromWChar(CatalogName, NameLength1, &CpLength1, &Stmt->Connection->charset, NULL);
-  CpSchema= MADB_ConvertFromWChar(SchemaName, NameLength2, &CpLength2, &Stmt->Connection->charset, NULL);
-  CpTable= MADB_ConvertFromWChar(TableName, NameLength3, &CpLength3, &Stmt->Connection->charset,NULL);
-  CpColumn= MADB_ConvertFromWChar(ColumnName, NameLength4, &CpLength4, &Stmt->Connection->charset, NULL);
+  CpSchema=  MADB_ConvertFromWChar(SchemaName,  NameLength2, &CpLength2, &Stmt->Connection->charset, NULL);
+  CpTable=   MADB_ConvertFromWChar(TableName,   NameLength3, &CpLength3, &Stmt->Connection->charset, NULL);
+  CpColumn=  MADB_ConvertFromWChar(ColumnName,  NameLength4, &CpLength4, &Stmt->Connection->charset, NULL);
 
   ret= Stmt->Methods->Columns(Stmt, CpCatalog, (SQLSMALLINT)CpLength1, CpSchema, (SQLSMALLINT)CpLength2,
                               CpTable, (SQLSMALLINT)CpLength3, CpColumn, (SQLSMALLINT)CpLength4);
@@ -690,12 +661,11 @@ SQLRETURN SQLConnectCommon(SQLHDBC ConnectionHandle,
   MDBUG_C_DUMP(Connection, Connection, 0x);
   MDBUG_C_DUMP(Connection, ServerName, s);
   MDBUG_C_DUMP(Connection, NameLength1, d);
-  MDBUG_C_DUMP(Connection,  UserName, s);
+  MDBUG_C_DUMP(Connection, UserName, s);
   MDBUG_C_DUMP(Connection, NameLength2, d);
   MDBUG_C_DUMP(Connection, Authentication, s);
   MDBUG_C_DUMP(Connection, NameLength3, d);
 
-  
   if (CheckConnection(Connection))
   {
     MADB_SetError(&Connection->Error, MADB_ERR_08002, NULL, 0);
@@ -724,8 +694,13 @@ SQLRETURN SQLConnectCommon(SQLHDBC ConnectionHandle,
 
   if (SQL_SUCCEEDED(ret))
   {
+    MADB_DSN_Free(Connection->Dsn);
     Connection->Dsn= Dsn;
   }
+  else
+  {
+    MADB_DSN_Free(Dsn);
+  }
 
   MDBUG_C_RETURN(Connection, ret, &Connection->Error);
 }
@@ -740,13 +715,6 @@ SQLRETURN SQL_API SQLConnect(SQLHDBC ConnectionHandle,
     SQLCHAR *Authentication,
     SQLSMALLINT NameLength3)
 {
-  MADB_Dbc *Dbc= (MADB_Dbc*)ConnectionHandle;
-  
-  if (!Dbc)
-    return SQL_INVALID_HANDLE;
-
-  MADB_CLEAR_ERROR(&Dbc->Error);
-
   return SQLConnectCommon(ConnectionHandle, ServerName, NameLength1,
                           UserName, NameLength2, Authentication, NameLength3);
 }
@@ -921,7 +889,7 @@ SQLRETURN SQL_API SQLDisconnect(SQLHDBC ConnectionHandle)
 {
   SQLRETURN ret= SQL_ERROR;
   MADB_Dbc *Connection = (MADB_Dbc *)ConnectionHandle;
-  LIST *Element, *NextElement;
+  MADB_List *Element, *NextElement;
 
   if (!Connection)
     return SQL_INVALID_HANDLE;
@@ -993,19 +961,19 @@ SQLRETURN SQL_API SQLDriverConnect(SQLHDBC ConnectionHandle,
 /* }}} */
 
 /* {{{ SQLDriverConnectW */
-SQLRETURN SQL_API SQLDriverConnectW(SQLHDBC ConnectionHandle,
-    SQLHWND WindowHandle,
-    SQLWCHAR *InConnectionString,
-    SQLSMALLINT StringLength1,
-    SQLWCHAR *OutConnectionString,
-    SQLSMALLINT BufferLength,
-    SQLSMALLINT *StringLength2Ptr,
-    SQLUSMALLINT DriverCompletion)
+SQLRETURN SQL_API SQLDriverConnectW(SQLHDBC      ConnectionHandle,
+                                    SQLHWND      WindowHandle,
+                                    SQLWCHAR    *InConnectionString,
+                                    SQLSMALLINT  StringLength1,
+                                    SQLWCHAR    *OutConnectionString,
+                                    SQLSMALLINT  BufferLength,
+                                    SQLSMALLINT *StringLength2Ptr,
+                                    SQLUSMALLINT DriverCompletion)
 {
   SQLRETURN   ret=          SQL_ERROR;
   SQLSMALLINT Length=       0;
   char        *InConnStrA=  NULL;
-  SQLULEN     StrLength=    0;
+  SQLULEN     InStrAOctLen= 0;
   char        *OutConnStrA= NULL;
   MADB_Dbc    *Dbc=         (MADB_Dbc *)ConnectionHandle;
    
@@ -1016,7 +984,7 @@ SQLRETURN SQL_API SQLDriverConnectW(SQLHDBC ConnectionHandle,
 
   MADB_CLEAR_ERROR(&Dbc->Error);
 
-  InConnStrA= MADB_ConvertFromWChar(InConnectionString, StringLength1, &StrLength, &utf8, NULL);
+  InConnStrA= MADB_ConvertFromWChar(InConnectionString, StringLength1, &InStrAOctLen, &utf8, NULL);
   MDBUG_C_DUMP(Dbc, Dbc, 0x);
   MDBUG_C_DUMP(Dbc, InConnStrA, s);
   MDBUG_C_DUMP(Dbc, StringLength1, d);
@@ -1028,32 +996,29 @@ SQLRETURN SQL_API SQLDriverConnectW(SQLHDBC ConnectionHandle,
   /* Allocate buffer for Asc OutConnectionString */
   if (OutConnectionString && BufferLength)
   {
-    Length= BufferLength;
+    Length= BufferLength*4 /*Max bytes per utf8 character */;
     OutConnStrA= (char *)MADB_CALLOC(Length);
+
+    if (OutConnStrA == NULL)
+    {
+      ret= MADB_SetError(&Dbc->Error, MADB_ERR_HY001, NULL, 0);
+      goto end;
+    }
   }
 
-  ret= Dbc->Methods->DriverConnect(Dbc, WindowHandle, (SQLCHAR *)InConnStrA, (SQLSMALLINT)StrLength, (SQLCHAR *)OutConnStrA,
+  ret= Dbc->Methods->DriverConnect(Dbc, WindowHandle, (SQLCHAR *)InConnStrA, (SQLSMALLINT)InStrAOctLen, (SQLCHAR *)OutConnStrA,
                                      Length, StringLength2Ptr, DriverCompletion); 
   MDBUG_C_DUMP(Dbc, ret, d);
   if (!SQL_SUCCEEDED(ret))
     goto end;
 
-  if (DriverCompletion == SQL_DRIVER_NOPROMPT &&
-      OutConnectionString && BufferLength)
+  if (OutConnectionString)
   {
     Length= (SQLSMALLINT)MADB_SetString(&utf8, OutConnectionString, BufferLength,
-                                      InConnStrA, StrLength, &((MADB_Dbc *)ConnectionHandle)->Error);
-    if (BufferLength < Length)
-      MADB_SetError(&Dbc->Error, MADB_ERR_01004, NULL, 0);
+                                        OutConnStrA, SQL_NTS, &((MADB_Dbc *)ConnectionHandle)->Error);
+    if (StringLength2Ptr)
+      *StringLength2Ptr= Length;
   }
-  if (DriverCompletion == SQL_DRIVER_COMPLETE ||
-      DriverCompletion == SQL_DRIVER_COMPLETE_REQUIRED)
-  {
-    Length= (SQLSMALLINT)MADB_SetString(&utf8, OutConnectionString, BufferLength,
-                                      OutConnStrA, SQL_NTS, &((MADB_Dbc *)ConnectionHandle)->Error);
-  }
-  if (StringLength2Ptr)
-    *StringLength2Ptr= Length;
   
 end:
   MADB_FREE(OutConnStrA);
@@ -1104,7 +1069,7 @@ SQLRETURN MA_SQLEndTran(SQLSMALLINT HandleType,
   case SQL_HANDLE_ENV:
     {
       MADB_Env *Env= (MADB_Env *)Handle;
-      LIST *List= Env->Dbcs;
+      MADB_List *List= Env->Dbcs;
 
       for (List= Env->Dbcs; List; List= List->next)
         ((MADB_Dbc *)List->data)->Methods->EndTran((MADB_Dbc *)List->data, CompletionType);
@@ -1296,33 +1261,20 @@ SQLRETURN SQL_API SQLTransact(SQLHENV Env, SQLHDBC Dbc, SQLUSMALLINT CompletionT
 /* }}} */
 
 /* {{{ SQLExecDirect */
-SQLRETURN MA_SQLExecDirect(MADB_Stmt *Stmt,
+SQLRETURN SQL_API SQLExecDirect(SQLHSTMT StatementHandle,
     SQLCHAR *StatementText,
     SQLINTEGER TextLength)
 {
+  MADB_Stmt *Stmt= (MADB_Stmt *)StatementHandle;
   SQLRETURN ret;
 
   if (!Stmt)
     ret= SQL_INVALID_HANDLE;
   else
     ret= Stmt->Methods->ExecDirect(Stmt, (char *)StatementText, TextLength);
+
   MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
 }
-
-
-SQLRETURN SQL_API SQLExecDirect(SQLHSTMT StatementHandle,
-    SQLCHAR *StatementText,
-    SQLINTEGER TextLength)
-{
-  MADB_Stmt *Stmt= (MADB_Stmt *)StatementHandle;
-
-  MDBUG_C_ENTER(Stmt->Connection, "SQLExecDirect");
-  MDBUG_C_DUMP(Stmt->Connection, StatementHandle, 0x);
-  MDBUG_C_DUMP(Stmt->Connection, StatementText, s);
-  MDBUG_C_DUMP(Stmt->Connection, TextLength, d);
-
-  return MA_SQLExecDirect(Stmt, StatementText, TextLength);
-}
 /* }}} */
 
 /* {{{ SQLExecDirectW */
@@ -1330,10 +1282,10 @@ SQLRETURN SQL_API SQLExecDirectW(SQLHSTMT StatementHandle,
     SQLWCHAR *StatementText,
     SQLINTEGER TextLength)
 {
-  char *CpStmt;
-  SQLULEN StmtLength;
+  char      *CpStmt;
+  SQLULEN   StmtLength;
   SQLRETURN ret;
-  BOOL ConversionError;
+  BOOL      ConversionError;
 
   MADB_Stmt *Stmt= (MADB_Stmt *)StatementHandle;
 
@@ -1353,23 +1305,13 @@ SQLRETURN SQL_API SQLExecDirectW(SQLHSTMT StatementHandle,
     ret= Stmt->Error.ReturnValue;
   }
   else
-    ret= Stmt->Methods->ExecDirect(Stmt, CpStmt, StmtLength);
+    ret= Stmt->Methods->ExecDirect(Stmt, CpStmt, (SQLINTEGER)StmtLength);
   MADB_FREE(CpStmt);
 
   MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
 }
 /* }}} */
 
-/* {{{ SQLExecute */
-SQLRETURN MA_SQLExecute(MADB_Stmt *Stmt)
-{
-  SQLRETURN ret;
-
-  ret= Stmt->Methods->Execute(Stmt);
-
-  MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
-}
-
 SQLRETURN SQL_API SQLExecute(SQLHSTMT StatementHandle)
 {
   MADB_Stmt *Stmt= (MADB_Stmt *)StatementHandle;
@@ -1382,7 +1324,7 @@ SQLRETURN SQL_API SQLExecute(SQLHSTMT StatementHandle)
   MDBUG_C_ENTER(Stmt->Connection, "SQLExecute");
   MDBUG_C_DUMP(Stmt->Connection, Stmt, 0x);
 
-  return MA_SQLExecute(Stmt);
+  return Stmt->Methods->Execute(Stmt, FALSE);
 }
 /* }}} */
 
@@ -1411,7 +1353,7 @@ SQLRETURN SQL_API SQLExtendedFetch(SQLHSTMT StatementHandle,
    
   Stmt->Ird->Header.RowsProcessedPtr= RowCountPtr;
   Stmt->Ird->Header.ArrayStatusPtr= RowStatusArray;
-  ret= MA_SQLFetchScroll((SQLHSTMT)Stmt, FetchOrientation, FetchOffset);
+  ret=  Stmt->Methods->FetchScroll(Stmt, FetchOrientation, FetchOffset);
 
   if (RowStatusArray && SaveArrayStatusPtr)
   {
@@ -1437,60 +1379,45 @@ SQLRETURN SQL_API SQLExtendedFetch(SQLHSTMT StatementHandle,
 /* }}} */
 
 /* {{{ SQLFetch */
-SQLRETURN MA_SQLFetch(SQLHSTMT StatementHandle)
-{
-  MADB_Stmt *Stmt= (MADB_Stmt *)StatementHandle;
-  SQLRETURN ret;
-  if (!Stmt)
-    return SQL_INVALID_HANDLE;
-  Stmt->FetchType= MADB_FETCH_TYPE_FETCH;
-  MDBUG_C_ENTER(Stmt->Connection, "SQLFetch");
-  ret= Stmt->Methods->Fetch(Stmt, FALSE);
-
-  MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
-}
-
 SQLRETURN SQL_API SQLFetch(SQLHSTMT StatementHandle)
 {
+  MADB_Stmt *Stmt;
+  
   if (StatementHandle == SQL_NULL_HSTMT)
     return SQL_INVALID_HANDLE;
-  MADB_CLEAR_ERROR(&((MADB_Stmt*)StatementHandle)->Error);
 
-  return MA_SQLFetch(StatementHandle);
+  Stmt= (MADB_Stmt *)StatementHandle;
+
+  MDBUG_C_ENTER(Stmt->Connection, "SQLFetch");
+  MADB_CLEAR_ERROR(&Stmt->Error);
+
+  /* SQLFetch is equivalent of SQLFetchScroll(SQL_FETCH_NEXT), 3rd parameter is ignored for SQL_FETCH_NEXT */
+  MDBUG_C_RETURN(Stmt->Connection, Stmt->Methods->FetchScroll(Stmt, SQL_FETCH_NEXT, 1), &Stmt->Error);
 }
 /* }}} */
 
 /* {{{ SQLFetchScroll */
-SQLRETURN MA_SQLFetchScroll(SQLHSTMT StatementHandle,
-    SQLSMALLINT FetchOrientation,
-    SQLLEN FetchOffset)
-{
-  MADB_Stmt *Stmt= (MADB_Stmt *)StatementHandle;
-  SQLRETURN ret;
-  if(!Stmt)
-    return SQL_INVALID_HANDLE;
-
-  MDBUG_C_ENTER(Stmt->Connection, "SQLFetchScroll");
-  MDBUG_C_DUMP(Stmt->Connection, FetchOrientation, d);
-
-  ret= Stmt->Methods->FetchScroll(Stmt, FetchOrientation, FetchOffset);
-
-  MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
-}
-
 SQLRETURN SQL_API SQLFetchScroll(SQLHSTMT StatementHandle,
     SQLSMALLINT FetchOrientation,
     SQLLEN FetchOffset)
 {
+  MADB_Stmt *Stmt;
+
   if (StatementHandle == SQL_NULL_HSTMT)
     return SQL_INVALID_HANDLE;
-  MADB_CLEAR_ERROR(&((MADB_Stmt*)StatementHandle)->Error);
 
-  return MA_SQLFetchScroll(StatementHandle, FetchOrientation, FetchOffset);
+  Stmt= (MADB_Stmt *)StatementHandle;
+
+  MDBUG_C_ENTER(Stmt->Connection, "SQLFetchScroll");
+  MDBUG_C_DUMP(Stmt->Connection, FetchOrientation, d);
+
+  MADB_CLEAR_ERROR(&Stmt->Error);
+
+  MDBUG_C_RETURN(Stmt->Connection, Stmt->Methods->FetchScroll(Stmt, FetchOrientation, FetchOffset), &Stmt->Error);
 }
 /* }}} */
 
-
+/* {{{ SQLFreeHandle */
 SQLRETURN SQL_API SQLFreeHandle(SQLSMALLINT HandleType,
                                 SQLHANDLE Handle)
 {
@@ -1515,7 +1442,8 @@ SQLRETURN SQL_API SQLFreeHandle(SQLSMALLINT HandleType,
       MDBUG_C_DUMP(Dbc, Handle, 0x);
 
       ret= MADB_DbcFree(Dbc);
-      MDBUG_C_RETURN(Dbc, ret, &Dbc->Error);
+      return ret;
+      /*MDBUG_C_RETURN(Dbc, ret, &Dbc->Error);*/
     }
   case SQL_HANDLE_DESC:
     {
@@ -1809,19 +1737,6 @@ SQLRETURN SQL_API SQLGetCursorNameW(
 /* }}} */
 
 /* {{{ SQLGetData */
-SQLRETURN MA_SQLGetData(SQLHSTMT StatementHandle,
-    SQLUSMALLINT Col_or_Param_Num,
-    SQLSMALLINT TargetType,
-    SQLPOINTER TargetValuePtr,
-    SQLLEN BufferLength,
-    SQLLEN *StrLen_or_IndPtr)
-{
-  MADB_Stmt *Stmt= (MADB_Stmt *)StatementHandle;
-  if (!Stmt)
-    return SQL_INVALID_HANDLE;
-  return Stmt->Methods->GetData(StatementHandle, Col_or_Param_Num, TargetType, TargetValuePtr, BufferLength, StrLen_or_IndPtr);
-}
-
 SQLRETURN SQL_API SQLGetData(SQLHSTMT StatementHandle,
     SQLUSMALLINT Col_or_Param_Num,
     SQLSMALLINT TargetType,
@@ -1829,11 +1744,53 @@ SQLRETURN SQL_API SQLGetData(SQLHSTMT StatementHandle,
     SQLLEN BufferLength,
     SQLLEN *StrLen_or_IndPtr)
 {
+  MADB_Stmt *Stmt= (MADB_Stmt*)StatementHandle;
+  unsigned int i;
+  MADB_DescRecord *IrdRec;
+
   if (StatementHandle== SQL_NULL_HSTMT)
     return SQL_INVALID_HANDLE;
-  MADB_CLEAR_ERROR(&((MADB_Stmt*)StatementHandle)->Error);
+  MADB_CLEAR_ERROR(&Stmt->Error);
 
-  return MA_SQLGetData(StatementHandle, Col_or_Param_Num, TargetType, TargetValuePtr, BufferLength, StrLen_or_IndPtr);
+  /* In case we don't have DM(it check for that) */
+  if (TargetValuePtr == NULL)
+  {
+    return MADB_SetError(&Stmt->Error, MADB_ERR_HY009, NULL, 0);
+  }
+
+  /* Bookmark */
+  if (Col_or_Param_Num == 0)
+  {
+    return MADB_GetBookmark(Stmt, TargetType, TargetValuePtr, BufferLength, StrLen_or_IndPtr);
+  }
+
+  /* We don't need this to be checked in case of "internal" use of the GetData, i.e. for internal needs we should always get the data */
+  if ( Stmt->CharOffset[Col_or_Param_Num - 1] > 0
+    && Stmt->CharOffset[Col_or_Param_Num - 1] >= Stmt->Lengths[Col_or_Param_Num - 1])
+  {
+    return SQL_NO_DATA;
+  }
+
+  if (BufferLength < 0)
+  {
+    return MADB_SetError(&Stmt->Error, MADB_ERR_HY090, NULL, 0);
+  }
+
+  /* reset offsets for other columns. Doing that here since "internal" calls should not do that */
+  for (i=0; i < mysql_stmt_field_count(Stmt->stmt); i++)
+  {
+    if (i != Col_or_Param_Num - 1)
+    {
+      IrdRec= MADB_DescGetInternalRecord(Stmt->Ird, i, MADB_DESC_READ);
+      if (IrdRec)
+      {
+        MADB_FREE(IrdRec->InternalBuffer);
+      }
+      Stmt->CharOffset[i]= 0;
+    }
+  }
+
+  return Stmt->Methods->GetData(StatementHandle, Col_or_Param_Num, TargetType, TargetValuePtr, BufferLength, StrLen_or_IndPtr, FALSE);
 }
 /* }}} */
 
@@ -2264,7 +2221,7 @@ SQLRETURN SQL_API SQLNativeSql(SQLHDBC ConnectionHandle,
     MADB_SetError(&Dbc->Error, MADB_ERR_01004, NULL, 0);
     return Dbc->Error.ReturnValue;
   }
-  Length= MADB_SetString(0, OutStatementText, BufferLength, (char *)InStatementText, TextLength1, &Dbc->Error);
+  Length= (SQLINTEGER)MADB_SetString(0, OutStatementText, BufferLength, (char *)InStatementText, TextLength1, &Dbc->Error);
   if (TextLength2Ptr)
     *TextLength2Ptr= Length;
   return Dbc->Error.ReturnValue;
@@ -2343,18 +2300,6 @@ SQLRETURN SQL_API SQLParamData(SQLHSTMT StatementHandle,
 }
 /* }}} */
 
-/* {{{ MA_SQLPrepare */
-SQLRETURN MA_SQLPrepare(MADB_Stmt *Stmt,
-    SQLCHAR *StatementText,
-    SQLINTEGER TextLength)
-{
-  SQLRETURN ret;
-  if (!Stmt)
-    return SQL_INVALID_HANDLE;
-  ret= Stmt->Methods->Prepare(Stmt, (char *)StatementText, TextLength);
-
-  MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
-}
 
 SQLRETURN SQL_API SQLPrepare(SQLHSTMT StatementHandle,
     SQLCHAR *StatementText,
@@ -2373,9 +2318,9 @@ SQLRETURN SQL_API SQLPrepare(SQLHSTMT StatementHandle,
   MDBUG_C_DUMP(Stmt->Connection, StatementText, s);
   MDBUG_C_DUMP(Stmt->Connection, TextLength, d);
 
-  MADB_CLEAR_ERROR(&Stmt->Error);
+  /* Prepare method clears error */
 
-  return MA_SQLPrepare(Stmt, StatementText, TextLength);
+  return Stmt->Methods->Prepare(Stmt, (char *)StatementText, TextLength, FALSE);
 }
 /* }}} */
 
@@ -2408,7 +2353,7 @@ SQLRETURN SQL_API SQLPrepareW(SQLHSTMT StatementHandle,
     ret= Stmt->Error.ReturnValue;
   }
   else
-    ret= Stmt->Methods->Prepare(Stmt, StmtStr, StmtLength);
+    ret= Stmt->Methods->Prepare(Stmt, StmtStr, (SQLINTEGER)StmtLength, FALSE);
   MADB_FREE(StmtStr);
 
   MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
@@ -2768,7 +2713,7 @@ SQLRETURN SQL_API SQLSetCursorNameW(SQLHSTMT StatementHandle,
   MADB_CLEAR_ERROR(&Stmt->Error);
 
   CpName= MADB_ConvertFromWChar(CursorName, NameLength, &Length, &Stmt->Connection->charset, NULL);
-  return Stmt->Methods->SetCursorName(Stmt, (char *)CpName, Length);
+  return Stmt->Methods->SetCursorName(Stmt, (char *)CpName, (SQLINTEGER)Length);
 }
 /* }}} */
 
@@ -2879,7 +2824,6 @@ SQLRETURN SQL_API SQLSetPos(SQLHSTMT StatementHandle,
   MDBUG_C_DUMP(Stmt->Connection, Operation, u);
   MDBUG_C_DUMP(Stmt->Connection, LockType, d);
 
-  Stmt->FetchType= MADB_FETCH_TYPE_SETPOS;
   ret= Stmt->Methods->SetPos(Stmt, RowNumber, Operation, LockType, 0);
 
   MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
@@ -2911,8 +2855,12 @@ SQLRETURN  SQL_API SQLBindParam(SQLHSTMT StatementHandle,
                                 SQLSMALLINT ParameterScale, SQLPOINTER ParameterValue,
                                 SQLLEN *StrLen_or_Ind)
 {
-  return SQLSetParam(StatementHandle, ParameterNumber, ValueType, ParameterType, LengthPrecision, ParameterScale,
-                      ParameterValue, StrLen_or_Ind);
+  if (!StatementHandle)
+    return SQL_INVALID_HANDLE;
+  MADB_CLEAR_ERROR(&((MADB_Stmt*)StatementHandle)->Error);
+
+  return MA_SQLBindParameter(StatementHandle, ParameterNumber, SQL_PARAM_INPUT_OUTPUT, ValueType, ParameterType, LengthPrecision, ParameterScale,
+                      ParameterValue, SQL_SETPARAM_VALUE_MAX, StrLen_or_Ind);
 
 }
 /* }}} */
@@ -3236,10 +3184,18 @@ SQLRETURN SQL_API SQLParamOptions(
     SQLULEN  *pirow)
 {
   MADB_Stmt *Stmt= (MADB_Stmt *)hstmt;
+  SQLRETURN result;
   if (!Stmt)
     return SQL_INVALID_HANDLE;
   MADB_CLEAR_ERROR(&Stmt->Error);
 
-  return MADB_DescSetField(Stmt->Apd, 0, SQL_DESC_ARRAY_SIZE, (SQLPOINTER)crow, SQL_IS_USMALLINT, 0);
+  result= MADB_DescSetField(Stmt->Apd, 0, SQL_DESC_ARRAY_SIZE, (SQLPOINTER)crow, SQL_IS_UINTEGER, 0);
+
+  if (SQL_SUCCEEDED(result))
+  {
+    result= MADB_DescSetField(Stmt->Ipd, 0, SQL_DESC_ROWS_PROCESSED_PTR, (SQLPOINTER)pirow, SQL_IS_POINTER, 0);
+  }
+
+  return result;
 }
 /* }}} */
